unit DragDropHandler;

// -----------------------------------------------------------------------------
//
//			*** NOT FOR RELEASE ***
//
// -----------------------------------------------------------------------------
// Project:         Drag and Drop Component Suite.
// Module:          DragDropHandler
// Description:     Implements Shell Drop Handlers.
// Version:         4.0
// Date:            25-JUN-2000
// Target:          Win32, Delphi 3-6 and C++ Builder 3-5
// Authors:         Angus Johnson, ajohnson@rpi.net.au
//                  Anders Melander, anders@melander.dk, http://www.melander.dk
// Copyright         1997-2000 Angus Johnson & Anders Melander
// -----------------------------------------------------------------------------
// Implements the following classes:
// * TCustomDropMultiTarget descendants:
//   - TDropHandler
// * TDropHandlerFactory
// -----------------------------------------------------------------------------
interface

uses
  DragDrop,
  DragDropFile,
  ActiveX,
  ComObj,
  Windows,
  Classes;

type
////////////////////////////////////////////////////////////////////////////////
//
//		TDropHandler
//
////////////////////////////////////////////////////////////////////////////////
// Based on Angus Johnson's DropHandler demo.
////////////////////////////////////////////////////////////////////////////////
// A typical drop handler session goes like this:
// 1. User drags one or more source files over a target file which has a
//    registered drop handler.
// 2. The shell loads the drop handler module.
// 3. The shell instantiates the registered drop handler object as an in-process
//    COM server.
// 4. The IPersistFile.Load method is called with the name of the target file.
//    The target file name is stored in the TDropHandler.TargetFile property.
// 5. The IDropTarget.Enter method is called. This causes a TDropHandler.OnEnter
//    event to be fired.
// 6. One of two things can happen next:
//    a) The user drops the source files on the target file.
//       The IDropTarget.Drop method is called. This causes a
//       TDropHandler.OnDrop event to be fired.
//       The names of the dropped files are stored in the TDropHandler.Files
//       string list property.
//    b) The user drags the source files away from the target file.
//       The IDropTarget.Leave method is called. This causes a
//       TDropHandler.OnLeave event to be fired.
// 7. The shell unloads the drop handler module (usually after a few seconds).
////////////////////////////////////////////////////////////////////////////////
  TDropHandler = class(TDropFileTarget, IPersistFile)
  private
    FTargetFile: string;
  protected
    // IPersistFile implementation
    function GetClassID(out classID: TCLSID): HResult; stdcall;
    function IsDirty: HResult; stdcall;
    function Load(pszFileName: POleStr; dwMode: Longint): HResult; stdcall;
    function Save(pszFileName: POleStr; fRemember: BOOL): HResult; stdcall;
    function SaveCompleted(pszFileName: POleStr): HResult; stdcall;
    function GetCurFile(out pszFileName: POleStr): HResult; stdcall;
  public
    property TargetFile: string read FTargetFile;
  end;

////////////////////////////////////////////////////////////////////////////////
//
//		TDropHandlerFactory
//
////////////////////////////////////////////////////////////////////////////////
  TDropHandlerFactory = class(TComObjectFactory)
  private
    FExtension: string;
    FName: string;
  protected
    function CreateInstance(const UnkOuter: IUnknown; const IID: TGUID;
      out Obj): HResult; stdcall;
  public
    constructor Create(ComServer: TComServerObject; ComponentClass: TComponentClass;
      const ClassID: TGUID; const ClassName, Description, AName,
      AExtension: string; Instancing: TClassInstancing);
    function CreateComObject(const Controller: IUnknown): TComObject; override;
    procedure UpdateRegistry(Register: Boolean); override;
    property Name: string read FName write FName;
    property Extension: string read FExtension write FExtension;
  end;

////////////////////////////////////////////////////////////////////////////////
//
//		Component registration
//
////////////////////////////////////////////////////////////////////////////////
procedure Register;


////////////////////////////////////////////////////////////////////////////////
//
//		Misc.
//
////////////////////////////////////////////////////////////////////////////////
procedure RegisterDropHandler(const Name, Extension, DropHandlerPath, ClassID: string);
procedure UnregisterDropHandler(const Name, Extension, ClassID: string);


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
//			IMPLEMENTATION
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
implementation

uses
  SysUtils;

////////////////////////////////////////////////////////////////////////////////
//
//		Component registration
//
////////////////////////////////////////////////////////////////////////////////

procedure Register;
begin
  RegisterComponents(DragDropComponentPalettePage, [TDropHandler]);
end;


////////////////////////////////////////////////////////////////////////////////
//
//		Utilities
//
////////////////////////////////////////////////////////////////////////////////
(*
HKEY_CLASSES_ROOT\CLSID\<ClassID>\InprocServer32
  Default: "<DropHandlerPath>"
  ThreadingModel: "Apartment"
HKEY_CLASSES_ROOT\<Extension>
  Default: "<Name>"
HKEY_CLASSES_ROOT\<Name>\shellex\DropHandler
  Default: "<ClassID>"
*)
procedure RegisterDropHandler(const Name, Extension, DropHandlerPath, ClassID: string);
begin
  CreateRegKey('CLSID\'+ClassID+'\InprocServer32', '', DropHandlerPath);
  CreateRegKey('CLSID\'+ClassID+'\InprocServer32', 'ThreadingModel', 'Apartment');
  CreateRegKey(Extension, '', Name);
  CreateRegKey(Name+'\shellex\DropHandler', '', ClassID);
end;

procedure UnregisterDropHandler(const Name, Extension, ClassID: string);
begin
  DeleteRegKey(Name+'\shellex\DropHandler');
  DeleteRegKey(Name+'\shellex');
  DeleteRegKey(Name);
  DeleteRegKey(Extension);
  DeleteRegKey('CLSID\'+ClassID+'\InprocServer32\ThreadingModel');
  DeleteRegKey('CLSID\'+ClassID+'\InprocServer32');
  DeleteRegKey('CLSID\'+ClassID);
end;


////////////////////////////////////////////////////////////////////////////////
//
//		TDropHandler
//
////////////////////////////////////////////////////////////////////////////////
function TDropHandler.GetClassID(out classID: TCLSID): HResult;
begin
  result := E_NOTIMPL;
end;

function TDropHandler.GetCurFile(out pszFileName: POleStr): HResult;
begin
  result := E_NOTIMPL;
end;

function TDropHandler.IsDirty: HResult;
begin
  result := S_FALSE;
end;

function TDropHandler.Load(pszFileName: POleStr; dwMode: Integer): HResult;
begin
  FTargetFile := WideCharToString(pszFileName);
  result := S_OK;
end;

function TDropHandler.Save(pszFileName: POleStr; fRemember: BOOL): HResult;
begin
  result := E_NOTIMPL;
end;

function TDropHandler.SaveCompleted(pszFileName: POleStr): HResult;
begin
  result := E_NOTIMPL;
end;


////////////////////////////////////////////////////////////////////////////////
//
//		TApartmentThread
//
////////////////////////////////////////////////////////////////////////////////
// Copied from VCLCom unit.
////////////////////////////////////////////////////////////////////////////////
type
  TApartmentThread = class(TThread)
  private
    FFactory: IClassFactory2;
    FUnkOuter: IUnknown;
    FIID: TGuid;
    FSemaphore: THandle;
    FStream: Pointer;
    FCreateResult: HResult;
  protected
    procedure Execute; override;
  public
    constructor Create(Factory: IClassFactory2; UnkOuter: IUnknown; IID: TGuid);
    destructor Destroy; override;
    property Semaphore: THandle read FSemaphore;
    property CreateResult: HResult read FCreateResult;
    property ObjStream: Pointer read FStream;
  end;

constructor TApartmentThread.Create(Factory: IClassFactory2;
  UnkOuter: IUnknown; IID: TGuid);
begin
  FFactory := Factory;
  FUnkOuter := UnkOuter;
  FIID := IID;
  FSemaphore := CreateSemaphore(nil, 0, 1, nil);
  FreeOnTerminate := True;
  inherited Create(False);
end;

destructor TApartmentThread.Destroy;
begin
  CloseHandle(FSemaphore);
  inherited Destroy;
end;

procedure TApartmentThread.Execute;
var
  msg: TMsg;
  Unk: IUnknown;
begin
  try
    CoInitialize(nil);
    try
      FCreateResult := FFactory.CreateInstanceLic(FUnkOuter, nil, FIID, '', Unk);
      FUnkOuter := nil;
      FFactory := nil;
      if FCreateResult = S_OK then
        CoMarshalInterThreadInterfaceInStream(FIID, Unk, IStream(FStream));
      ReleaseSemaphore(FSemaphore, 1, nil);
      if FCreateResult = S_OK then
        while GetMessage(msg, 0, 0, 0) do
        begin
          DispatchMessage(msg);
          Unk._AddRef;
          if Unk._Release = 1 then break;
        end;
    finally
      Unk := nil;
      CoUninitialize;
    end;
  except
    { No exceptions should go unhandled }
  end;
end;

////////////////////////////////////////////////////////////////////////////////
//
//		TVCLComObject
//
////////////////////////////////////////////////////////////////////////////////
// Based on TVCLAutoObject
////////////////////////////////////////////////////////////////////////////////
type
  TVCLComObject = class(TComObject, IVCLComObject, IUnknown)
  private
    FComponent: TComponent;
    FOwnsComponent: Boolean;
  protected
    // IVCLComObject implementation
    procedure FreeOnRelease;
    function Invoke(DispID: Integer; const IID: TGUID;
      LocaleID: Integer; Flags: Word; var Params;
      VarResult, ExcepInfo, ArgErr: Pointer): HResult;  virtual; stdcall;
    function GetIDsOfNames(const IID: TGUID; Names: Pointer;
      NameCount, LocaleID: Integer; DispIDs: Pointer): HResult; virtual; stdcall;
    function GetTypeInfo(Index, LocaleID: Integer; out TypeInfo): HResult; virtual; stdcall;
    function GetTypeInfoCount(out Count: Integer): HResult; virtual; stdcall;
  public
    // TODO : For now, please ignore linker warning about TVCLComObject.Create
    constructor Create(Factory: TComObjectFactory; Component: TComponent);
    destructor Destroy; override;
    procedure Initialize; override;
    function ObjQueryInterface(const IID: TGUID; out Obj): HResult; override;
  end;

constructor TVCLComObject.Create(Factory: TComObjectFactory;
  Component: TComponent);
begin
  FComponent := Component;
  CreateFromFactory(Factory, nil);
end;

destructor TVCLComObject.Destroy;
begin
  if FComponent <> nil then
  begin
    FComponent.VCLComObject := nil;
    if FOwnsComponent then
      FComponent.Free;
  end;
  inherited Destroy;
end;

procedure TVCLComObject.FreeOnRelease;
begin
  FOwnsComponent := True;
end;

function TVCLComObject.GetIDsOfNames(const IID: TGUID; Names: Pointer;
  NameCount, LocaleID: Integer; DispIDs: Pointer): HResult;
begin
  Result := E_NOTIMPL;
end;

function TVCLComObject.GetTypeInfo(Index, LocaleID: Integer;
  out TypeInfo): HResult;
begin
  Pointer(TypeInfo) := nil;
  Result := E_NOTIMPL;
end;

function TVCLComObject.GetTypeInfoCount(out Count: Integer): HResult;
begin
  Count := 0;
  Result := E_NOTIMPL;
end;

procedure TVCLComObject.Initialize;
begin
  inherited Initialize;
  if FComponent = nil then
  begin
    FComponent := TComponentClass(Factory.ComClass).Create(nil);
    FOwnsComponent := True;
  end;
  FComponent.VCLComObject := Pointer(IVCLComObject(Self));
end;

function TVCLComObject.Invoke(DispID: Integer; const IID: TGUID;
  LocaleID: Integer; Flags: Word; var Params; VarResult, ExcepInfo,
  ArgErr: Pointer): HResult;
begin
  Result := E_NOTIMPL;
end;

function TVCLComObject.ObjQueryInterface(const IID: TGUID;
  out Obj): HResult;
begin
  Result := inherited ObjQueryInterface(IID, Obj);
  if (Result <> 0) and (FComponent <> nil) then
    if FComponent.GetInterface(IID, Obj) then
      Result := 0;
end;

////////////////////////////////////////////////////////////////////////////////
//
//		TDropHandlerFactory
//
////////////////////////////////////////////////////////////////////////////////
constructor TDropHandlerFactory.Create(ComServer: TComServerObject;
  ComponentClass: TComponentClass; const ClassID: TGUID; const ClassName,
  Description, AName, AExtension: string; Instancing: TClassInstancing);
begin
  inherited Create(ComServer, TComClass(ComponentClass), ClassID, ClassName,
    Description, Instancing, tmApartment);
  FName := AName;
  FExtension := AExtension;
end;

function TDropHandlerFactory.CreateComObject(const Controller: IUnknown): TComObject;
begin
  Result := TVCLComObject.CreateFromFactory(Self, Controller);
end;

function TDropHandlerFactory.CreateInstance(const UnkOuter: IUnknown;
  const IID: TGUID; out Obj): HResult;
begin
  if not IsLibrary and (ThreadingModel = tmApartment) then
  begin
    LockServer(True);
    try
      with TApartmentThread.Create(Self, UnkOuter, IID) do
      begin
        if WaitForSingleObject(Semaphore, INFINITE) = WAIT_OBJECT_0 then
        begin
          Result := CreateResult;
          if Result <> S_OK then Exit;
          Result := CoGetInterfaceAndReleaseStream(IStream(ObjStream), IID, Obj);
        end else
          Result := E_FAIL
      end;
    finally
      LockServer(False);
    end;
  end else
    Result := inherited CreateInstance(UnkOuter, IID, Obj);
end;

type
  TComponentProtectedAccess = class(TComponent);
  TComponentProtectedAccessClass = class of TComponentProtectedAccess;

procedure TDropHandlerFactory.UpdateRegistry(Register: Boolean);

  function GetModuleFileName: string;
  var
    Buffer: array[0..261] of Char;
  begin
    SetString(Result, Buffer, Windows.GetModuleFileName(HInstance,
      Buffer, SizeOf(Buffer)));
  end;

begin
  if Register then
  begin
    inherited UpdateRegistry(Register);
    CreateRegKey(Extension, '', Name);
    CreateRegKey(Name+'\shellex\DropHandler', '', GUIDToString(ClassID));
  end;
  TComponentProtectedAccessClass(ComClass).UpdateRegistry(Register,
    GUIDToString(ClassID), ProgID);
  if not Register then
  begin
    RegDeleteKey(HKEY_CLASSES_ROOT, PChar(Name+'\shellex\DropHandler'));
    RegDeleteKey(HKEY_CLASSES_ROOT, PChar(Name+'\shellex'));
    RegDeleteKey(HKEY_CLASSES_ROOT, PChar(Name));
    RegDeleteKey(HKEY_CLASSES_ROOT, PChar(Extension));
    inherited UpdateRegistry(Register);
  end;
end;

end.
